from tqdm import tqdm
from ..data_manager import DataManager
from sklearn.metrics import multilabel_confusion_matrix

from ....utils_vvd import ListOrderedDict


class EvaluatorBase(object):
    TYPE = None
    def __init__(self, data_gt, data_pred, verbose=True):
        """
        :classname_list: list of classes that we are interested in
        """
        self.verbose = verbose

        # load datamanager
        self._dm_gt = self._load_dm_obj(data_gt)
        self._dm_pred = self._load_dm_obj(data_pred)

        # get am obj from dm
        self.am = self._get_am(self._dm_gt, self._dm_pred)
        self.classnames = self.am.classnames

        # check datamanager obj before evaluation
        self._eva_check(self._dm_gt, self._dm_pred)
        self.gt_data_list, self.pred_data_list, self.data_path_list = self.extract_data_pair(self._dm_gt, self._dm_pred)

    def extract_data_pair(self, dm_gt, dm_pred):
        self._logging(f"@@ extracting data pair.")
        _gt_data_dict = self._get_gt_result_dict(dm_gt)
        _pred_data_dict = self._get_pred_result_dict(dm_pred)
        _data_path_dict = self._get_data_path_dict(dm_gt)

        assert isinstance(_gt_data_dict, dict), f"{_gt_data_dict} should be an obj of dict"
        assert isinstance(_pred_data_dict, dict), f"{_pred_data_dict} should be an obj of dict"
        assert isinstance(_data_path_dict, dict), f"{_data_path_dict} should be an obj of dict"
        assert len(dm_gt) == len(dm_pred) == len(_gt_data_dict) == len(_pred_data_dict) == len(_data_path_dict), f"data num error : {len(dm_gt)} {len(dm_pred)} {len(_gt_data_dict)} {len(_pred_data_dict)} {len(_data_path_dict)}"
        assert set(_gt_data_dict) == set(_pred_data_dict) == set(_data_path_dict), f"uuid keys of dm objs do not match."

        self._logging(f"@@ total data {len(dm_gt)}, start packing data pair.")
        gt_data_list = list()
        pred_data_list = list()
        data_path_list = list()
        for key in tqdm(list(_gt_data_dict), desc="Packing data: "):
            gt_data_list.append(_gt_data_dict[key])
            pred_data_list.append(_pred_data_dict[key])
            data_path_list.append(_data_path_dict[key])

        return gt_data_list, pred_data_list, data_path_list

    @staticmethod
    def _get_am(dm_gt, dm_pred):
        assert dm_gt.am._class_dict == dm_pred.am._class_dict, f"class_dicts are not same. {dm_gt.am._class_dict}, {dm_pred.am._class_dict}"
        assert type(dm_gt.am) == type(dm_pred.am), f"AM data type {type(dm_gt.am)}, {dm_pred.am}"
        assert dm_gt.am.classnames == dm_pred.am.classnames, f"{dm_gt.am.classnames} != {dm_pred.am.classnames}"
        return dm_gt.am

    def _load_dm_obj(self, data):
        if isinstance(data, str):
            return DataManager.load(data)
        elif isinstance(data, DataManager):
            return data
        else:
            raise RuntimeError("Unknown data format: {}".format(type(data)))

    def _logging(self, data=None):
        if data is None:
            return getattr(self, 'verbose', False)
        else:
            if getattr(self, 'verbose', False):
                print(data)

    def _eva_check(self, data_gt, data_pred):
        # type check
        assert type(data_gt) == type(data_pred), f"type of data_gt {type(data_gt)} and data_pred {type(data_pred)} are not same."
        if 'DMML' in str(type(data_gt)):
            assert self.TYPE == 'MultiLabel', f"data's type {str(type(data_gt))} is not same as evaluator's {self.TYPE}"
        if 'DMMC' in str(type(data_gt)):
            assert self.TYPE == 'MultiClass', f"data's type {str(type(data_gt))} is not same as evaluator's {self.TYPE}"
        if 'DMDet' in str(type(data_gt)):
            assert self.TYPE == 'Detection', f"data's type {str(type(data_gt))} is not same as evaluator's {self.TYPE}"
        self._logging("@@ Checking input data.")
        assert isinstance(data_gt, DataManager), f"data_gt {data_gt} should be an obj of DataManager"
        assert isinstance(data_pred, DataManager), f"data_gt {data_pred} should be an obj of DataManager"
        assert data_gt and data_pred, "Please load both gt and pred before evaluation."
        assert data_gt == data_pred, "Make sure the gt data and pred data match"
        self._data_check(data_gt, data_pred)

    def set_threshold(self, policy='inflection', value=None, *args, **kwargs):
        """
        Args:
            policy (str, optional): could be one of 'inflection', 'recall', 'precision' or 'manual'. Defaults to 'inflection'.
            value (optional): 
                under 'inflection' policy: value is insignificant
                under 'recall' or 'precision': value should be the minimum float we can accept
                under 'manual': value can be a list or dict who has the same length or keys with classnames. 
                . Defaults to None.
        """
        thre_dict = self._set_threshold(policy=policy, value=value, *args, **kwargs)
        assert len(thre_dict) == len(self.classnames), f"data length error {len(thre_dict)} != {len(self.classnames)}"
        assert set(thre_dict) == set(self.classnames), f"duplicate key error {set(thre_dict)} != {set(self.classnames)}"
        for _, value in thre_dict.items():
            assert 0 <= value <= 1, f"bad threshold value {value} in {thre_dict}"
        threshold_dict = ListOrderedDict()
        for key in self.classnames:
            threshold_dict[key] = thre_dict[key]
        self.threshold_dict = threshold_dict

    def _get_data_path_dict(self, data):
        data_path_dict = dict()
        for rec in data:
            uuid = rec['info']['uuid']
            data_path = rec['info']['image_path']
            assert uuid not in data_path_dict, f"duplicate uuid in dataset {uuid}"
            data_path_dict[uuid] = data_path
        return data_path_dict

    def _get_gt_result_dict(self, dm_gt):
        raise NotImplementedError(f"Base Evaluator func, _get_gt_result_dict, not Implemented")

    def _get_pred_result_dict(self, dm_pred):
        raise NotImplementedError(f"Base Evaluator func, _get_pred_result_dict, not Implemented")

    def _data_check(self, gt_data, pd_data):
        raise NotImplementedError(f"Base Evaluator func, _data_check, not Implemented")

    def eval_ap(self, *args, **kwargs):
        raise NotImplementedError(f"Base Evaluator func, _eval_ap, not Implemented")

    def eval_judgment(self, *args, **kwargs):
        raise NotImplementedError(f"Base Evaluator func, _eval_ap, not Implemented")
    
    def _set_threshold(self, *args, **kwargs):
        raise NotImplementedError(f"Base Evaluator func, _set_threshold, not Implemented")

    def get_confusion_matrix(self):
        raise NotImplementedError(f"Base Evaluator func, get_confusion_matrix, not Implemented")
    
    def dump_failure_case(self, data_root, target_dir):
        raise NotImplementedError(f"Base Evaluator func, dump_failure_case, not Implemented")



from pathlib import Path
import os
import uuid
import platform

def scan_folder(data_dir, extension='bmp'):
    """
    Generate a dictionary, where key: file name, and value: file path
    Assuming all file names are unique in data_dir
    param:
        data_dir: directory where you want to scan the file, recursively
        extension: the file extension. this param specifies the file format
    return:
        list of paths to files with the given extension
    """
    assert os.path.exists(data_dir), data_dir+" does not exist"
    extension = extension.strip('.')
    files = Path(data_dir).glob('**/*.{}'.format(extension))
    files = [str(x) for x in files]
    return files


def create_tmp_file(extension=''):
    """
    Return a temporary path that is randomly generated.
    Arguments:
        extension: the extension of the generated path (no dot)
    Return:
        a randomly generated path (str) in /tmp folder
    """
    if platform.system() == 'Windows':
        tmp_folder = r'./TempFiles'
    else:
        tmp_folder = '/tmp/temp-random-files/'
    
    if extension != '':
        tmp_file = os.path.join(tmp_folder, uuid.uuid1().hex) + '.' + extension.lstrip('.')
    else:
        tmp_file = os.path.join(tmp_folder, uuid.uuid1().hex)

    return tmp_file



# -*- coding: utf-8 -*-
# @Author: Zhai Menghua
# @Date:   2020-07-27 15:16:16
# @Last Modified by:   Zhai Menghua
# @Last Modified time: 2020-07-31 12:38:42
from .id_related import *
from .path_related import *
from .coding_related import *
from .mean_ap import eval_map
from .classifier_eval import ClassifierEvalBinary, ClassifierEvalMulticlass, ClassifierEvalMultilabel
from .eva_utils import get_TFPN



from ..utils import decode_distribution, decode_labelme_shape
from ..utils import encode_labelme_shape
from ..utils import create_uuid
import numpy as np
from .base import AMBase
from .utils import logging


class AMDet(AMBase):
    def __init__(self, *args, **kwargs):
        super(AMDet, self).__init__(*args, **kwargs)
        self.set_class()


    def set_class(self, arg='leaf'):
        """
        Set classes of interested (default 'all')
        """
        if None: pass
        elif arg == 'leaf':
            classname_list = self._leaf_classes
        elif isinstance(arg, (list, tuple)):
            classname_list = arg
        else:
            raise RuntimeError("Unknown args")

        self._check_classes_conflict(classname_list)

        self._class2label = {name:label for label, name in enumerate(classname_list)}
        self._label2class = {label:name for label, name in enumerate(classname_list)}

        # global class ids of interested classes
        self._classid_of_interested = [self._class2label_all[classname] for classname in classname_list]
        self._distribution_classes = classname_list


    #@logging
    def get_score(self, arg):
        if None: pass
        elif isinstance(arg, dict):
            if 'instances' in arg:  # a record
                return [self.get_score(inst) for inst in arg['instances']]
            else:                   # a instance
                inst = arg
                if 'class_id' in inst or 'class_name' in inst:
                    assert 'score' in inst
                    return inst['score']
                elif 'distribution' in inst:
                    dist_all = decode_distribution(inst['distribution'])
                    dist_oin = [dist_all[ix] for ix in self._classid_of_interested]
                    return float(np.max(dist_oin))
                else:
                    raise RuntimeError("inst: ", inst)
        else:
            raise TypeError("Argument Type Not Supported: {}".format(type(arg)))

    #@logging
    def get_shape(self, arg):
        if None: pass
        elif isinstance(arg, dict):
            if 'instances' in arg:  # a record
                return [self.get_shape(inst) for inst in arg['instances']]
            else:                   # a instance
                inst = arg
                assert 'points' in inst
                return decode_labelme_shape(inst['points'])
        else:
            raise TypeError("Argument Type Not Supported: {}".format(type(arg)))

    #@logging
    def get_shape(self, arg):
        if None: pass
        elif isinstance(arg, dict):
            if 'instances' in arg:  # a record
                return [self.get_shape(inst) for inst in arg['instances']]
            else:                   # a instance
                inst = arg
                assert 'points' in inst
                return decode_labelme_shape(inst['points'])
        else:
            raise TypeError("Argument Type Not Supported: {}".format(type(arg)))

    #@logging
    def get_xyxy(self, arg):
        points = self.get_shape(arg)
        def _points2xyxy(points):
            points = np.reshape(points, [-1,2])
            x1,y1,x2,y2 = points[:,0].min(), points[:,1].min(), points[:,0].max(), points[:,1].max()
            return [x1,y1,x2,y2]
        if len(points) == 0:
            return []
        elif isinstance(points[0], (list, tuple)):
            return [_points2xyxy(pts) for pts in points]
        else:
            return _points2xyxy(points)

    
    #@logging
    def get_xywh(self, arg):
        points = self.get_shape(arg)
        def _points2xywh(points):
            points = np.reshape(points, [-1,2])
            x1,y1,x2,y2 = points[:,0].min(), points[:,1].min(), points[:,0].max(), points[:,1].max()
            return [x1,y1,x2-x1,y2-y1]
        if len(points) == 0:
            return []
        elif isinstance(points[0], (list, tuple)):
            return [_points2xywh(pts) for pts in points]
        else:
            return _points2xywh(points)


    #@logging
    def create_instance(self, classname, score=1, shape=[]):
        """
        return a bbox or a list of bboxes
        """
        # create class id
        assert classname in self.classnames


        # create shape
        shape_type = 'polygon'
        assert isinstance(shape, (list, tuple, np.ndarray))
        if None: pass
        elif len(shape) == 0:
            shape_type = None
            points = []
        elif len(shape) == 2:
            assert isinstance(shape[0], (list, tuple, np.ndarray))
            (x1,y1),(x2,y2) = shape[0], shape[1]
            points = [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y2]]
        elif len(shape) == 4:
            mark_list = [not hasattr(item, '__iter__') for item in shape]
            if all(mark_list):
                x1, y1, x2, y2 = shape
                points = [[x1, y1], [x1, y2], [x2, y2], [x2, y1], [x1, y2]]
            elif not any(mark_list):
                points = [[x, y] for x, y in shape]
            else:
                raise RuntimeError("Invalid shape: {}".format(shape))
        elif len(shape) >= 3:
            points = [[x, y] for x, y in shape]
        else:
            raise RuntimeError("Invalid shape: {}".format(shape))

        instance = {
            'uuid': create_uuid(),
            'class_name': classname,
            'score': score,
            'shape_type': shape_type,
            'points': encode_labelme_shape(points)
        } 

        return instance


#
# Annotation manager
#
from .utils import decode_distribution
from .utils import decode_labelme_shape
from .utils import encode_distribution
from .utils import encode_labelme_shape
from .utils import create_uuid
import numpy as np

def logging(func):
    def wrapper(*args, **kw):
        res = func(*args, **kw)
        print("----------------")
        print("{}():".format(func.__name__))
        print("Inputs:")
        for idx, arg in enumerate(args[1:]):
            print("\targ{} = {}".format(idx+1, arg))
        for key, val in kw.items():
            print("\t{} = {}".format(key, val))
        print("Return: \n\t{}".format(res))
        print("----------------\n\n")
        return res
    return wrapper


class AnnotationManager(object):
    def __init__(self, class_dict=[]):
        """ class_dict could be get from the data manager object """
        print("WARNING: class {} is depreciated, use AMDet/AMML/AMMC instead".format(str(self)))
        assert isinstance(class_dict, (list, tuple))
        self._class_dict = sorted(class_dict, key=lambda x: x['class_id'])
        self._class2label_global = {x['class_name']:x['class_id'] for x in class_dict}
        self._label2class_global = {x['class_id']:x['class_name'] for x in class_dict}

        self._all_classes = [x['class_name'] for x in self._class_dict]
        self._leaf_classes = self.__get_leaf_classes(class_dict)
        self._OK_classname = 'OK'  # class name for OK (in DataManager, we don't save OK as a class in distribution)

        # initialize class object
        self.setup_distribution_type(distribution_type='detection', distribution_classes=self._leaf_classes)

    @property
    def classnames(self):
        return self._distribution_classes

    def __get_leaf_classes(self, class_dict):
        # 1) collect all parent classnames
        nonleaf_classes = set([x['parent'] for x in class_dict if x['parent'] is not None])
        leaf_classes = list()
        for x in self._class_dict:
            if x['class_name'] not in nonleaf_classes:
                leaf_classes.append(x['class_name'])
        return leaf_classes

    def __check_classes_conflict(self, classname_list):
        """ because we have parent classes, this function checks if there are parent class
            and its child class in the classname_list.
        """
        # loop over classname_list to check if one's parent is also in classname_list
        for classname in classname_list:
            if classname == self._OK_classname:
                continue
            parent = 'WHAT THE FUCK'
            child  = classname
            while parent is not None:
                classid_global = self._class2label_global[child]
                parent = self._class_dict[classid_global]['parent']
                if parent in classname_list:
                    raise RuntimeError("(WARNING) conflict class pair found: ({}, {})".format(child, parent))
                child = parent

    # @logging
    def setup_distribution_type(self, distribution_type=None, distribution_classes=None):
        """
        :distribution_type: string, there're 3 modes: 'multilabel' (default), 'multiclass' and 'detection'
        :distribution_classes: list of strings, the output distribution that indicates the relations between class names and labels
        """
        if distribution_classes is not None:
            assert isinstance(distribution_classes, (list, tuple))
        else:
            distribution_classes = self._distribution_classes

        if distribution_type is None:
            distribution_type = self._distribution_type

        if None: pass
        elif distribution_type in ['multilabel', 'multi-label']:
            self._distribution_type = 'multilabel'
            self._distribution_classes = distribution_classes

        elif distribution_type in ['multiclass', 'multi-class', 'seg', 'segmentation']:
            self.__check_classes_conflict(distribution_classes)
            self._distribution_type = 'multiclass'
            self._distribution_classes = distribution_classes

        elif distribution_type in ['detection', 'det']:
            self.__check_classes_conflict(distribution_classes)
            self._distribution_type = 'detection'
            self._distribution_classes = distribution_classes

        else:
            raise RuntimeError('Unknown distribution type: {}'.format(distribution_type))

        self._class2label = {name:label for label, name in enumerate(self._distribution_classes)}
        self._label2class = {label:name for label, name in enumerate(self._distribution_classes)}


    def __guess_object_type(self, obj):
        """
        get_xxx() method series take an ambiguous object as their major input.
        this function returns a type guess about the object for them, by checking their type
        """
        exception = RuntimeError("Illegible object type ({}): {}".format(type(obj), obj))
        if isinstance(obj, str):
            return 'classname'
        elif isinstance(obj, (int, np.int32, np.int64)):
            return 'classid'
        elif isinstance(obj, dict):
            if 'instances' in obj:
                return 'record'
            else:
                return 'instance'
        elif isinstance(obj, (tuple, list)):
            if len(obj) == 0:
                return 'empty-list'
            elif isinstance(obj[0], (tuple, list)):
                return 'list-list'
            elif isinstance(obj[0], str):
                return 'string-list'
            elif isinstance(obj[0], (float, int)):
                return 'number-list'
            elif isinstance(obj[0], dict):
                key_type = self.__guess_object_type(obj[0])
                return '{}-list'.format(key_type)
            else:
                raise exception
        else:
            raise exception


    # @logging
    def get_classid(self, obj):
        """
        :obj: could be a distribution, record, instance of list, instances or a class name
        return [] (if no instance or matched class name), or class id in current distribution type
        """
        obj_type = self.__guess_object_type(obj)
        if obj_type == 'number-list':    # we assume this is a distribution
            distribution = obj
            return [class_id for class_id, score in enumerate(distribution) if score > 0]
        elif obj_type == 'list-list':    # we assume this is a list of distribution
            distribution_list = obj
            return [self.get_classid(distribution) for distribution in distribution_list]
        else:
            distributions = self.get_distribution(obj)
            assert len(distributions) > 0
            if isinstance(distributions[0], list):  # if returns a list of distribution
                return [self.get_classid(distribution) for distribution in distributions]
            else:
                distribution = distributions
                return self.get_classid(distribution)

    
    # @logging
    def get_classname(self, obj):
        """
        :obj: could be a distribution, record, instance of list, instances or a class label (in distribution classes format)
        return [] (if no instance or matched class name), or class id in current distribution type
        """
        eps = 1e-8
        obj_type = self.__guess_object_type(obj)
        if None: pass
        elif obj_type == 'classid':
            classid = obj
            return self._label2class[classid]
        elif obj_type == 'number-list':    # we assume this is a distribution
            distribution = np.array(obj)
            assert len(distribution) == len(self._distribution_classes)
            if self._distribution_type == 'multilabel':
                return [self.get_classname(classid) for classid, score in enumerate(distribution) if score > 0]
            else:
                max_score = np.max(distribution)
                return [self.get_classname(classid) for classid, score in enumerate(distribution) if (score == max_score) and (max_score > eps)]
        elif obj_type == 'list-list':     # we assume this is a list of distribution
            distribution_list = obj
            return [self.get_classname(distribution) for distribution in distribution_list]
        elif obj_type == 'empty-list':    # we assume this is from the empty instance of a OK record: rec['instances']
            if self._distribution_type == 'detection':
                return []
            else:
                return [self._OK_classname]
        else:
            distributions = self.get_distribution(obj)
            assert len(distributions) > 0
            if isinstance(distributions[0], list):  # if returns a list of distribution
                return [self.get_classname(distribution) for distribution in distributions]
            else:
                distribution = distributions
                return self.get_classname(distribution)


    # @logging
    def get_score(self, obj):
        """
        :obj: could be a record, distribution (list), a instance (list)
        return [] (if no instance or matched class name), or class id in current distribution type
        """
        obj_type = self.__guess_object_type(obj)
        if obj_type == 'number-list':    # we assume this is a distribution
            distribution = obj
            return [score for class_id, score in enumerate(distribution) if score > 0]
        elif obj_type == 'list-list':    # we assume this is a list of distribution
            distribution_list = obj
            return [self.get_score(distribution) for distribution in distribution_list]
        else:
            distributions = self.get_distribution(obj)
            return self.get_score(distributions)


    # @logging
    def get_distribution(self, obj):
        """
        :obj: could be a record, instance of list, instances
        return a probability distribution defined by distribution classes
        """
        obj_type = self.__guess_object_type(obj)
        if None: pass
        elif obj_type == 'classname':
            classname = obj
            distribution = [0] * len(self._distribution_classes)
            distribution[self._class2label[classname]] = 1
            return distribution
        elif obj_type == 'string-list':  # we assume this is a classname list
            classname_list = obj
            distribution = [0] * len(self._distribution_classes)
            for classname in classname_list:
                distribution[self._class2label[classname]] = 1
            return distribution
        elif obj_type == 'empty-list':
            return [1 if x == self._OK_classname else 0 for x in self._distribution_classes]
        elif obj_type == 'record':
            record = obj
            return self.get_distribution(record['instances'])
        elif obj_type == 'instance':
            instance = obj
            distribution_global = decode_distribution(instance['distribution'])
            distribution_target = list()
            ok_score = None
            for classname in self._distribution_classes:
                if classname == self._OK_classname:
                    ok_score = score = 0
                else:
                    score = distribution_global[self._class2label_global[classname]]
                distribution_target.append(score)
            if ok_score is not None:
                assert ok_score == 0
                ok_score = max(0, 1-sum(distribution_target))
                ok_id = self._class2label[self._OK_classname]
                distribution_target[ok_id] = ok_score
            return distribution_target
        elif obj_type == 'instance-list':
            instances = obj
            distributions = [self.get_distribution(inst) for inst in instances]
            if self._distribution_type in ['multiclass', 'multilabel']:  # in classification mode, squeeze to a single distribution
                distribution = np.max(distributions, axis=0)
                return distribution.tolist()
            else:
                return distributions
        else:
            raise RuntimeError("Unknown type: {} is a {}".format(obj, obj_type))


    # @logging
    def get_xyxy(self, obj):
        """
        :obj: could be a record, instance of list, instances
        return a bbox or a list of bboxes, or None if shape_type = None
        """
        obj_type = self.__guess_object_type(obj)
        if obj_type == 'empty-list':
            return []
        elif obj_type == 'instance-list':
            instances = obj
            return [self.get_xyxy(inst) for inst in instances]
        elif obj_type == 'record':    # we assume this is a list of distribution
            record = obj
            return self.get_xyxy(record['instances'])
        elif obj_type == 'instance':
            instance = obj
            if instance['shape_type'] is None:
                print("Warning: you are trying to extract bboxes from classification predictions/annotation")
                return None
            else:
                points = decode_labelme_shape(instance['points'])
                points = np.reshape(points, [-1,2])
                x1,y1,x2,y2 = points[:,0].min(), points[:,1].min(), points[:,0].max(), points[:,1].max()
                return [x1,y1,x2,y2]
        elif obj_type == 'number-list':
            if len(obj) == 4:
                # is a xyxy
                return obj
            else:
                raise RuntimeError("bad bumber_list: {}".format(obj))
        else:
            raise RuntimeError("Unknown type: {}".format(obj))

    
    # @logging
    def get_xywh(self, obj):
        """
        :obj: could be a record, instance of list, instances
        return a bbox or a list of bboxes, or None if shape_type = None
        """
        xyxy = self.get_xyxy(obj)
        if None: pass
        elif xyxy is None or len(xyxy) == 0:
            return xyxy
        elif isinstance(xyxy[0], list):
            return [self.get_xywh(bbox) for bbox in xyxy]
        else:
            x1,y1,x2,y2 = xyxy
            return [x1,y1,x2-x1,y2-y1]


    # @logging
    def get_polygon(self, obj):
        """
        :obj: could be a record, instance of list, instances
        return a bbox or a list of bboxes
        """
        obj_type = self.__guess_object_type(obj)
        if obj_type == 'empty-list':
            return []
        elif obj_type == 'instance-list':
            instances = obj
            return [self.get_polygon(inst) for inst in instances]
        elif obj_type == 'record':    # we assume this is a list of distribution
            record = obj
            return self.get_polygon(record['instances'])
        elif obj_type == 'instance':
            instance = obj
            if None: pass
            elif instance['shape_type'] is None:
                print("Warning: you are trying to extract bboxes from classification predictions/annotation")
                return None
            elif instance['shape_type'] == 'polygon':
                points = decode_labelme_shape(instance['points'])
                points = np.reshape(points, [-1,2])
                return points.tolist()
            elif instance['shape_type'] == 'rectangle':
                x1,y1,x2,y2 = decode_labelme_shape(instance['points'])
                return [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y1]]
            else:
                raise RuntimeError("Unknown shape type: {}".format(instance['shape_type']))
        else:
            raise RuntimeError("Unknown type: {}".format(obj))


    def _extend_distribution(self, distribution_global, exclude_classes=[]):
        """ extend the global distribution: fill parents with their maximal child """
        for child_id, child_score in enumerate(distribution_global):
            # trace back to root class
            parent = self._class_dict[child_id]['parent']
            while parent is not None:
                parent_id = self._class2label_global[parent]
                parent_score = distribution_global[parent_id]
                if parent not in exclude_classes:
                    distribution_global[parent_id] = max(child_score, parent_score)  # update parent score
                parent = self._class_dict[parent_id]['parent']   # update parent
                child_score = max(child_score, parent_score)     # update child_score
        return distribution_global


    # @logging
    def create_instance(self, obj, score=1, shape=[]):
        """
        :obj: could be a distribution (list) and class_name
        return a bbox or a list of bboxes
        """
        # create distribution
        if None: pass
        elif isinstance(obj, (list, np.ndarray)):
            distribution = list(obj)
            assert len(distribution) == len(self._distribution_classes)
            distribution_global = [0]*len(self._all_classes)
            for class_id, sc in enumerate(distribution):
                class_name = self._label2class[class_id]
                if class_name == self._OK_classname:  # skip ok probability
                    continue
                class_id_global = self._class2label_global[class_name]
                distribution_global[class_id_global] = sc
                distribution_global = self._extend_distribution(distribution_global, exclude_classes=self._distribution_classes)
        elif isinstance(obj, str):
            class_name = obj
            distribution_global = [0]*len(self._all_classes)
            if class_name != self._OK_classname:  # skip ok probability
                class_id_global = self._class2label_global[class_name]
                distribution_global[class_id_global] = score
                distribution_global = self._extend_distribution(distribution_global, exclude_classes=[class_name])
        else:
            raise RuntimeError("Unknown type: {}".format(obj))


        # create shape
        shape_type = 'polygon'
        assert isinstance(shape, (list, tuple, np.ndarray))
        if None: pass
        elif len(shape) == 0:
            assert self._distribution_type != 'detection'
            shape_type = None
            points = []
        elif len(shape) == 2:
            assert isinstance(shape[0], (list, tuple, np.ndarray))
            (x1,y1),(x2,y2) = shape[0], shape[1]
            points = [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y2]]
        elif len(shape) == 4:
            # assert isinstance(shape[0], (int, float, np.float32))
            mark_list = [not hasattr(item, '__iter__') for item in shape]
            if all(mark_list):
                x1, y1, x2, y2 = shape
                points = [[x1, y1], [x1, y2], [x2, y2], [x2, y1], [x1, y2]]
            elif not any(mark_list):
                points = [[x, y] for x, y in shape]
            else:
                raise RuntimeError("Invalid shape: {}".format(shape))
        elif len(shape) >= 3:
            points = [[x, y] for x, y in shape]
        else:
            raise RuntimeError("Invalid shape: {}".format(shape))

        instance = {
            'uuid': create_uuid(),
            'distribution': encode_distribution(distribution_global),
            'shape_type': shape_type,
            'points': encode_labelme_shape(points)
        } 

        return instance

    def show(self, image, obj, thickness=2, visual=True):
        """
        [show annotations on input image]
        Args:
            image ([numpy]): [input image]
            obj ([record, [instance], instance]): [could be a record, instance of list, instances]
        """

        import cv2
        color_list = [(159, 2, 98), (95, 32, 219), (222, 92, 189), (56, 233, 120), (23, 180, 100), (78, 69, 20), (97, 202, 39), (65, 179, 135), (163, 159, 219)]

        image = image.copy()

        old_dist_type = self._distribution_type
        self.setup_distribution_type(distribution_type='detection')
        polygon_list = self.get_polygon(obj)
        classname_list = self.get_classname(obj)
        self.setup_distribution_type(distribution_type=old_dist_type)

        obj_type = self.__guess_object_type(obj)

        if obj_type == 'instance':
            polygon_list = [polygon_list]
            classname_list = [classname_list]

        for polygon, class_name in zip(polygon_list, classname_list):
            class_id = self.get_classid(class_name)[-1]
            color = color_list[class_id % len(color_list)][1]
            # cv2.fillPoly(image, [np.array(polygon).astype('int32')], color)
            cv2.polylines(image, [np.array(polygon).astype('int32')], True, color, thickness=thickness)

        if visual:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.imshow(image)
            plt.show()

        return image



from .lib.data_manager import DataManager
from .lib.anno_manager import AnnotationManager
from .lib.evaluator import ClassificationEvaluator, ClassificationMultiLabelEvaluator, MultiLabelEvaluator
from .lib.evaluator import ClassificationMultiClassEvaluator, MultiClassEvaluator
from .lib.evaluator import DetectionEvaluator
from .lib.utils import *
from .lib.labelme import *
from .lib.labelme_utils import *
from .lib.anno import *
from .lib.dm import DMDet, DMMC, DMML



from ..utils import decode_distribution, encode_distribution
from .base import AMBase
from .utils import logging
import numpy as np


class AMMC(AMBase):
    def __init__(self, *args, **kwargs):
        super(AMMC, self).__init__(*args, **kwargs)
        self.set_class()


    def set_class(self, arg='leaf'):
        """
        Set classes of interested (default 'all')
        """
        if None: pass
        elif arg == 'leaf':
            classname_list = self._leaf_classes
        elif isinstance(arg, (list, tuple)):
            classname_list = arg
        else:
            raise RuntimeError("Unknown args")

        self._check_classes_conflict(classname_list)
        self._class2label = {name:label for label, name in enumerate(classname_list)}
        self._label2class = {label:name for label, name in enumerate(classname_list)}

        # global class ids of interested classes
        self._classid_of_interested = [self._class2label_all[classname] for classname in classname_list]
        self._distribution_classes = classname_list


    #@logging
    def get_classid(self, arg):
        if isinstance(arg, dict):
            rec = arg
            if 'info' in rec:
                info = rec['info']
                if 'distribution' in info:
                    dist_all = decode_distribution(info['distribution'])
                    dist_oin = [dist_all[ix] for ix in self._classid_of_interested]
                    return int(np.argmax(dist_oin))
                else:
                    return RuntimeError()
            else:
                raise RuntimeError()
        else:
            return super(AMMC, self).get_classid(arg)

    #@logging
    def get_score(self, arg):
        if None: pass
        elif isinstance(arg, dict):
            if 'info' in arg:
                rec = arg
                if 'distribution' in rec['info']:
                    dist_all = decode_distribution(rec['info']['distribution'])
                    dist_oin = [dist_all[ix] for ix in self._classid_of_interested]
                    return max(dist_oin)
                else:
                    raise RuntimeError('Input:', arg)
            else:
                raise RuntimeError()
        else:
            raise TypeError("Argument Type Not Supported: {}".format(type(arg)))

    #@logging
    def create_distribution(self, obj, score=1):
        """
        obj:
            - classname (str): classname + score in effect
            - distribution: distribution of classes of interested
        return:
            encoded scores (full distribution)
        """
        # create distribution
        if None: pass
        elif isinstance(obj, (tuple, list, np.ndarray)):
            distribution = list(obj)
            assert len(distribution) == len(self.classnames)
            distribution_all = [0]*len(self._all_classes)
            for class_id, sc in enumerate(distribution):
                class_name = self._label2class[class_id]
                class_id_global = self._class2label_all[class_name]
                distribution_all[class_id_global] = sc
                distribution_all = self._extend_distribution(distribution_all, exclude_classes=self.classnames)
        elif isinstance(obj, str):
            class_name = obj
            distribution_all = [0]*len(self._all_classes)
            class_id_global = self._class2label_all[class_name]
            distribution_all[class_id_global] = score
            distribution_all = self._extend_distribution(distribution_all, exclude_classes=[class_name])
        else:
            raise RuntimeError("Unknown type: {}".format(obj))

        return encode_distribution(distribution_all)


from tqdm import tqdm

from sklearn.metrics import confusion_matrix
from pathlib import Path
from ....utils_vvd import encode_path
from ....utils_vvd import cv_rgb_imwrite
from ....utils_vvd import cv_rgb_imread
from .base import EvaluatorBase
import pandas as pd


class ClassificationMultiClassEvaluator(EvaluatorBase):
    TYPE = 'MultiClass'
    def _data_check(self, gt_data, pd_data):
        assert gt_data.class_dict == pd_data.class_dict, f"differect class_dict between {gt_data.class_dict} and {pd_data.class_dict}"
        pass

    def _get_result_dict(self, dm):
        result_dict = dict()
        for rec in dm:
            class_name = self.am.get_classname(rec)
            result_dict[rec['info']['uuid']] = class_name
        return result_dict

    def _get_gt_result_dict(self, dm_gt):
        return self._get_result_dict(dm_gt)

    def _get_pred_result_dict(self, dm_pred):
        return self._get_result_dict(dm_pred)

    def get_confusion_matrix(self):
        y_true = self.gt_data_list
        y_pred = self.pred_data_list
        cm = confusion_matrix(y_true=y_true, y_pred=y_pred, labels=self.classnames)
        cm_pd = pd.DataFrame(cm, index=self.classnames, columns=self.classnames)
        return cm_pd
    
    def dump_failure_case(self, data_root, target_dir):
        target_dir = Path(target_dir)
        for y_true, y_pred, data_path in tqdm(zip(self.gt_data_list, self.pred_data_list, self.data_path_list), total=len(self.gt_data_list)):
            if y_true != y_pred:
                ori_img_path = Path(data_root) / data_path
                if not ori_img_path.exists():
                    print(f"Ori image path {ori_img_path} not found.")
                    continue
                else:
                    try:
                        image = cv_rgb_imread(ori_img_path)
                    except Exception as e:
                        print(f"image load failed! {ori_img_path}")
                        print(f"error message: {e}")
                        continue
                gt_label = self.classnames[y_true]
                pred_label = self.classnames[y_pred]
                save_path = target_dir / gt_label / pred_label / encode_path(ori_img_path)
                cv_rgb_imwrite(image, save_path)


MultiClassEvaluator = ClassificationMultiClassEvaluator



from email.policy import default
from .data_manager import DataManager as DataManagerBase

from .utils import decode_distribution, encode_distribution
from .anno import AMDet
from .anno import AMML
from .anno import AMMC

import numpy as np
import copy

def logging(func):
    def wrapper(*args, **kw):
        res = func(*args, **kw)
        print("----------------")
        print("{}():".format(func.__name__))
        print("Inputs:")
        for idx, arg in enumerate(args[1:]):
            print("\targ{} = {}".format(idx+1, arg))
        for key, val in kw.items():
            print("\t{} = {}".format(key, val))
        print("Return: \n\t{}".format(res))
        print("----------------\n\n")
        return res
    return wrapper


class DMBase(DataManagerBase):
    def __init__(self, record_list, class_dict):
        assert isinstance(record_list, (list, tuple))
        self.record_list = record_list
        self.class_dict = class_dict


class DMMC(DMBase):
    def __init__(self, *args, **kwargs):
        super(DMMC, self).__init__(*args, **kwargs)
        self.am = AMMC(self.class_dict)
        self.set_class = self.am.set_class
        self.get_classid = self.am.get_classid
        self.get_classname = self.am.get_classname
        self.get_score = self.am.get_score
        self.create_distribution = self.am.create_distribution

    def data_statistics(self):
        class_occurrence = self.occurrence(self.get_classname)
        return {
            'image_number': len(self),
            'class_occurrence': class_occurrence
        }


class DMML(DMBase):
    def __init__(self, *args, **kwargs):
        super(DMML, self).__init__(*args, **kwargs)
        self.am = AMML(self.class_dict)
        self.set_class = self.am.set_class
        self.get_classid = self.am.get_classid
        self.get_classname = self.am.get_classname
        self.get_score = self.am.get_score
        self.create_scores = self.am.create_scores


    def to_mc(self):
        # add OK class to class dict
        class_dict_aug = self.am._add_ok_class(self.class_dict)
        
        # transform record list
        am = self.am.clone()
        am.set_class('leaf')

        am_aug = AMMC(class_dict_aug)
        am_aug.set_class('leaf')

        record_list = list()
        for rec in self.record_list:
            new_rec = copy.deepcopy(rec['info'])
            new_rec.pop('scores')
            scores_leaf = am.get_score(rec)
            leaf_max_id = np.argmax(scores_leaf)
            ng_score = scores_leaf[leaf_max_id]
            leaf_class_with_max_score = am._label2class[leaf_max_id]
            all_max_id = am._class2label_all[leaf_class_with_max_score]
            distribution_tail = [0]*len(am._class2label_all)
            distribution_tail[all_max_id] = ng_score
            distribution = [1.-ng_score] + distribution_tail
            distribution = am_aug._extend_distribution(distribution)
            new_rec['distribution'] = encode_distribution(distribution)
            record_list.append(
                {
                    'info': new_rec
                }
            )

        return DMMC(record_list = record_list, class_dict = class_dict_aug)


    def data_statistics(self):
        return {'image_number': len(self)}


class DMDet(DMBase):
    def __init__(self, *args, **kwargs):
        super(DMDet, self).__init__(*args, **kwargs)
        self.am = AMDet(self.class_dict)
        self.set_class = self.am.set_class
        self.get_classid = self.am.get_classid
        self.get_classname = self.am.get_classname
        self.get_score = self.am.get_score
        self.get_shape = self.am.get_shape
        self.create_instance = self.am.create_instance

    # @logging
    def to_ml(self):
        # transform record list
        num_classes = len(self.class_dict)
        am = self.am
        record_list = list()
        for rec in self.record_list:
            new_rec = copy.deepcopy(rec['info'])
            distributions = [[0]*num_classes]
            for inst in rec['instances']:
                if 'class_id' in inst:
                    assert 'score' in inst
                    dist_all = [0]*num_classes
                    dist_all[inst['class_id']] = inst['score']
                    dist_all = am._extend_distribution(dist_all)
                elif 'class_name' in inst:
                    class_name = inst['class_name']
                    full_class_id = am._class2label_all[class_name]
                    assert 'score' in inst
                    dist_all = [0]*num_classes
                    dist_all[full_class_id] = inst['score']
                    dist_all = am._extend_distribution(dist_all)
                else:
                    dist_all = decode_distribution(inst['distribution'])
                distributions.append(dist_all)

            scores = np.max(distributions, axis=0)
            new_rec['scores'] = encode_distribution(scores)
            record_list.append(
                {
                    'info': new_rec
                }
            )

        return DMML(record_list = record_list, class_dict = self.class_dict)

    def to_mc(self):
        data_ml = self.to_ml()
        return data_ml.to_mc()


    def data_statistics(self):
        am = self.am.clone()
        am.set_class('leaf')

        class_occurence_by_instance = self.occurrence(am.get_classname)
        class_occurence_by_image = self.occurrence(lambda rec: list(set(am.get_classname(rec))))
        ok_ng_counter = self.occurrence(lambda rec: len(rec['instances'])>0)

        return {
            'total_images': len(self),
            'ok_images': ok_ng_counter.get(False, 0),
            'ng_images': ok_ng_counter.get(True,  0),
            'class_occurence_by_image': class_occurence_by_image,
            'class_occurence_by_instance': class_occurence_by_instance
        }


import numpy as np
from numpy.lib.function_base import iterable
from .utils import current_system, get_list_from_list
from .utils import is_number


def is_box(box):
    """
    check if the input box is a box
    """
    def is_iter_box(iter_box):
        if len(iter_box) != 4:
            return False
        else:
            for num in iter_box:
                if not is_number(num):
                    return False
        return True

    if isinstance(box, list):
        return is_iter_box(box)
    elif isinstance(box, tuple):
        return is_iter_box(box)
    elif isinstance(box, np.ndarray):
        return is_iter_box(box)
    else:
        return False


def compute_box_area(box):
    """
    compute area of input box
    """
    assert is_box(box)
    area = max(0, box[2] - box[0]) * max(0, box[3] - box[1])
    return area


def compute_box_box_iou(box1, box2):
    """
    compute iou of input boxes
    """
    area1 = compute_box_area(box1)
    area2 = compute_box_area(box2)
    u_x_min = max(box1[0], box2[0])
    u_x_max = min(box1[2], box2[2])
    u_y_min = max(box1[1], box2[1])
    u_y_max = min(box1[3], box2[3])
    new_box = [u_x_min, u_y_min, u_x_max, u_y_max]
    u_area = compute_box_area(new_box)
    iou = u_area / max(area1 + area2 - u_area, 1)
    return iou


def get_xyxy(polygon_xy):
    """
    get xyxy of a ploygon
    """
    polygon_array = np.array(polygon_xy)
    assert polygon_array.shape[0] > 1
    assert polygon_array.shape[1] == 2
    x1, y1, x2, y2 = polygon_array[:, 0].min(), polygon_array[:, 1].min(), polygon_array[:, 0].max(), polygon_array[:, 1].max()
    return [x1, y1, x2, y2]


def make_box(center_point, box_x, box_y=None):
    """ build box for a given center-point"""
    box_x = int(box_x)
    if box_y is None:
        box_y = box_x
    else:
        box_y = int(box_y)
    assert box_x > 0 and box_y > 0
    center_x, center_y = center_point

    left = int(round(center_x - box_x // 2))
    right = left + box_x
    top = int(round(center_y - box_y // 2))
    bottom = top + box_y

    box = [left, top, right, bottom]
    return box


def boxes_painter(rgb_image, box_list, label_list=None, score_list=None, color_list=None, color=None, class_name_dict=None, line_thickness=3):
    """[paint boxex and labels on image]

    Args:
        rgb_image ([np.array(uint8)]): [np array image as type uint8]
        box_list ([list of list of 4 int]): [list of box like [10(xmin), 20(ymin), 50(xmax), 60(ymax)]]
        label_list ([list of int]): [class indexes of boxes in box_list] (could be none)
        class_name_dict ([dict - index: class_name]): [key is index and value is the name in type of str] (could be none)
    Returns:
        [rgb image]: [image with boxes and labels]
    """

    if rgb_image.ndim == 2:
        rgb_image = (np.repeat(rgb_image[:, :, None], 3, axis=2))

    H, W = rgb_image.shape[:2]
    rgb_image = rgb_image.astype('uint8')

    color_input = color

    if label_list is not None:
        assert len(label_list) == len(box_list)
        if class_name_dict is not None:
            if isinstance(class_name_dict, dict):
                for item in label_list:
                    assert item in class_name_dict
            else:
                assert iterable(class_name_dict)
                assert len(class_name_dict) >= len(set(label_list))

    if score_list is not None:
        assert len(score_list) == len(box_list)

    if color_list is not None:
        assert len(color_list) == len(box_list)

    from PIL import ImageFont, ImageDraw, Image
    import matplotlib.font_manager as fm

    color_list_default = [(159, 2, 98), (95, 32, 219), (222, 92, 189), (56, 233, 120), (23, 180, 100), (78, 69, 20), (97, 202, 39), (65, 179, 135), (163, 159, 219)]

    pil_image = Image.fromarray(rgb_image)
    draw = ImageDraw.Draw(pil_image)

    fontsize = 24

    try:
        if current_system() == 'Windows':
            font = ImageFont.truetype('arial.ttf', fontsize)
        else:
            font = ImageFont.truetype(fm.findfont(fm.FontProperties(family='DejaVu Sans')),fontsize)
    except IOError:
        font = ImageFont.load_default()

    text_height = 22

    # draw boxes
    for index, bbox in enumerate(box_list):
        if not bbox:
            continue

        if color_input is not None:
            color = tuple(color_input)
        else:
            if color_list is not None:
                color = color_list[index]
            else:
                if label_list:
                    try:
                        color = color_list_default[label_list[index] % len(color_list_default)]
                    except:
                        color = (255, 255, 0)
                else:
                    color = (255, 255, 0)

        # draw text
        display_str = ""

        if label_list:
            if class_name_dict:
                display_str += class_name_dict[label_list[index]]
            else:
                display_str += str(label_list[index])

        if score_list:
            if display_str != "":
                display_str += ' '
            score = score_list[index]
            display_str += str(format(score, '.3f'))

        text_width, text_height = font.getsize(display_str)
        margin = np.ceil(0.05 * text_height)

        array_box = np.array(bbox)
        if array_box.ndim == 1:
            # box 
            assert len(bbox) == 4
            left, top, right, bottom = np.array(bbox).astype('int').tolist()
            points = [(left, top), (left, bottom), (right, bottom), (right, top), (left, top)]

        elif array_box.ndim == 2:
            # polygon
            assert array_box.shape[1] == 2
            bbox.append(bbox[0])
            points = bbox
            top_index = np.argmin(array_box[:, 1])
            left, top = bbox[top_index]
            right = text_width + left + margin * 2
        else:
            raise RuntimeError(f'unknown bbox shape {bbox}')

        points = get_list_from_list(points, lambda x: tuple(x))
        draw.line(points, width=line_thickness, fill=tuple(color))

        text_bottom = top
        
        ori_text_box = [left - 1, text_bottom - text_height - 2 * margin, right + 1, text_bottom]
        w_offset = h_offset = 0
        if ori_text_box[0] < 0:
            w_offset -= ori_text_box[0]
        if ori_text_box[2] > W:
            w_offset += W - ori_text_box[2]
        if ori_text_box[1] < 0:
            h_offset -= ori_text_box[1]
        if ori_text_box[3] > H:
            h_offset += H - ori_text_box[3]
        text_box = (np.array(ori_text_box) + [w_offset, h_offset] * 2).tolist()

        draw.rectangle(text_box, fill=tuple(color))
        # if np.mean(np.array(color)) < 250:
        #     font_color = 'yellow'
        # else:
        #     font_color = 'red'
        font_color = tuple((255 - np.array(color)).tolist())

        draw.text((int(text_box[0] + (text_box[2] - text_box[0])/2 - text_width/2), text_box[3] - text_height - 2 * margin), display_str, fill=font_color, font=font)

    # get image with box and index
    array_image_with_box = np.asarray(pil_image)

    return array_image_with_box

def adjust_bbox(bbox, offset_list):
    """
    [top, right, bottom, left]
    [A] -> [-A, -A, A, A]
    [A, B] -> [-A, -B, A, B]
    [A, B, C, D] -< [A, B, C, D]
    """
    assert is_box(bbox), f"bbox {bbox} is not a box."
    if is_number(offset_list):
        offset_list = [offset_list]
    
    if len(offset_list) == 1:
        A = offset_list[0]
        offset_list = [-A, -A, A, A]
    elif len(offset_list) == 2:
        A, B = offset_list
        offset_list = [-A, -B, A, B]
    elif len(offset_list) == 4:
        pass
    else:
        raise RuntimeError(f"unknown value {offset_list}")
    
    new_box = [
        bbox[0] + offset_list[0],
        bbox[1] + offset_list[1],
        bbox[2] + offset_list[2],
        bbox[3] + offset_list[3]
    ]
    
    return new_box



def logging(func):
    def wrapper(*args, **kw):
        res = func(*args, **kw)
        print("----------------")
        print("{}():".format(func.__name__))
        print("Inputs:")
        for idx, arg in enumerate(args[1:]):
            print("\targ{} = {}".format(idx+1, arg))
        for key, val in kw.items():
            print("\t{} = {}".format(key, val))
        print("Return: \n\t{}".format(res))
        print("----------------\n\n")
        return res
    return wrapper


from lib.dm import DMDet
import numpy as np

data_det = DMDet.load('assets/toy.json')
data_det.set_class('leaf')  # set classes of interested


for rec in data_det:
    if len(rec['instances']):
        inst = rec['instances'][0]
    else:
        inst = rec

    # get class id
    data_det.get_classid(rec)      # return a list of integers
    data_det.get_classid(inst)     # return 1 integer
    data_det.get_classid('dirt')   # return 1 integer
    data_det.get_classid(['scratch', 'grindmark'])  # return a list of integers

    # get class name
    data_det.get_classname(rec)   # return a list of strings
    data_det.get_classname(inst)  # return 1 string
    data_det.get_classname(3)     # return 1 string
    data_det.get_classname([0, 3, 1])  # return a list of strings (of class 0, 3, 1, in order)

    # get score
    data_det.get_score(rec)  # return a list of score (same length with get_classid(rec))
    data_det.get_score(inst) # return the instance score

    # get shape
    data_det.get_shape(rec)  # return a list of shapes
    data_det.get_shape(inst) # return 1 shape


data_ml = data_det.to_ml()

for rec in data_ml:
    # get class id
    data_ml.get_classid('dent')            # return 1 integer
    data_ml.get_classid(['dent', 'dent'])  # return a list of integers

    # get class name
    data_ml.get_classname(3)          # return 1 string
    data_ml.get_classname([0, 3, 1])  # return a list of strings (of class 0, 3, 1, in order)

    # get score
    data_ml.get_score(rec)    # return scores of all classes





data_mc = data_det.to_mc()

for rec in data_mc:
    # get class id
    data_mc.get_classid(rec)        # return the class id with max score
    data_mc.get_classid('scratch')  # return 1 integer
    data_mc.get_classid(['dent', 'grindmark'])  # return a list of integers

    # get class name
    data_mc.get_classname(rec)   # return the class name with max score
    data_mc.get_classname(0)     # return 1 string
    data_mc.get_classname([0, 3, 1])  # return a list of strings (of class 0, 3, 1, in order)

    # get score
    data_mc.get_score(rec)  # return the max score


data_ml2mc = data_ml.to_mc()  # ml2mc, pick the class with highest score and suppress the other classes,
                              # then and add a OK class to make the distribution sum up to one (class dict chaIBes)


data_ml2mc.set_class(['OK', 'NG'])
data_ml2mc.get_score(rec)
data_ml2mc.get_classname(rec)


data_det.create_instance('scratch', score=.7, shape=[100,20,221,35])

data_ml.create_scores('scratch', score=.3)
data_ml.create_scores(np.random.rand(len(data_ml.am.classnames)))

data_mc.create_distribution('OK', score=.3)
data_mc.create_distribution(np.random.rand(len(data_mc.am.classnames)))

data_det.save('/tmp/det.json')
data_ml.save('/tmp/ml.json')
data_mc.save('/tmp/mc.json')



from multiprocessing import Pool
from multiprocessing import Manager
import time
import numpy as np
from tqdm import tqdm
from .utils import vvd_round

from concurrent.futures import ThreadPoolExecutor


def single_process(main_fun, paras, share_list, share_lock, total_size, start_time, catch_exception):

    if catch_exception:
        try:
            res = main_fun(paras)
        except Exception as e:
            print(e)
            return
    else:
        res = main_fun(paras)
    if res is not None:
        share_lock.acquire()
        share_list.append(res)
        share_lock.release()
    else:
        print("None res got, and it won't appear in the share_list. ")

    print(f"{len(share_list)} / {total_size}")
    
    min_time = (time.time() - start_time) / max(1, len(share_list)) * (total_size - max(1, len(share_list))) / 60
    print(f"time left {min_time} min")


def multi_process(main_fun, para_fun, total_size, max_pool_num=8, catch_exception=True):
    start_time = time.time()
    if max_pool_num > 0:
        mp_manager = Manager()
        share_list = mp_manager.list()
        share_lock = mp_manager.Lock()
        pool = Pool(max_pool_num)
        for index in range(total_size):
            paras = para_fun(index)
            pool.apply_async(func=single_process, args=(main_fun, paras, share_list, share_lock, total_size, start_time, catch_exception))
        pool.close()
        pool.join()
        result_list = list(share_list)
    else:
        result_list = list()
        for index in range(total_size):
            res = main_fun(para_fun(index))
            result_list.append(res)
    print(f"mean time {(time.time() - start_time) / max(1, len(result_list))}")
    return result_list


class MultiProcessManager:
    TaskModeList = ['multiprocessing', 'multithreading', 'sequence']

    def __init__(self, core_function, running_mode='sequence', max_worker_num=None):
        self.core_function = core_function
        assert running_mode in type(self).TaskModeList, f"running_mode {running_mode} must in {type(self).TaskModeList}."
        self.running_mode = running_mode
        if max_worker_num is None:
            self.max_worker_num = 16
        else:
            self.max_worker_num = max(1, vvd_round(max_worker_num))

        if self.running_mode == 'multiprocessing':
            self.process_pool = Pool(self.max_worker_num)
        elif self.running_mode == 'multithreading':
            self.thread_pool = ThreadPoolExecutor(max_workers=self.max_worker_num)

    def infer(self, input_list):
        """
            "Pool.map results are ordered. If you need order, great; if you don't,
            Pool.imap_unordered may be a useful optimization. Note that while the
            order in which you receive the results from Pool.map is fixed, the
            order in which they are computed is arbitrary."
        -- https://stackoverflow.com/questions/41273960/python-3-does-pool-keep-the-original-order-of-data-passed-to-map
        """

        num_tasks = len(input_list)

        if None: pass
        elif self.running_mode == 'multiprocessing':
            # * multiprocessing
            pool = self.process_pool
            result_list = pool.map(self.core_function, input_list)
            pool.close()
            pool.join()

        elif self.running_mode == 'multithreading':
            # * multi threading
            result_list = list(self.thread_pool.map(self.core_function, input_list))

        elif self.running_mode == 'sequence':
            # * sequence
            result_list = list(map(self.core_function, input_list))
        else:
            raise TypeError

        assert isinstance(result_list, list), f"result_list {result_list} must be a list"

        # check the output order
        if len(result_list) != num_tasks:
            raise RuntimeError(f"Error: Something fatal happened during inference!!!!\
                len(result_list) {len(result_list)} != len(result_list) {len(result_list)}")

        return result_list



if __name__ == '__main__':
    def para_gen(index):
        para_1 = np.random.randint(0, 200)
        para_2 = np.random.randint(0, 200)
        return {'para_1':para_1, 'para_2':para_2}

    def main_fun_v(para_dict):
        para_1 = para_dict['para_1']
        para_2 = para_dict['para_2']
        return para_1 + para_2

    res_list = multi_process(main_fun_v, para_gen, 200, 0)
    pass


from labelme import Labelme
import cv2

def main():
    labelme1 = Labelme.from_json('assets/sample1.json')
    labelme2 = Labelme.from_json('assets/sample2.json')

    labelme_same = labelme1.intersection(labelme2)
    labelme_diff = labelme1.symmetric_difference(labelme2)

    labelme_same.save_json('tmp/same.json')
    # labelme_diff.save_json('tmp/diff.json')

    image = labelme_diff.load_image(image_root='/home/zhai/Documents/MatrixTime/Huatian/data/')
    image = labelme_diff.draw_shapes(image)
    cv2.imwrite('tmp/diff.jpg', image[...,::-1])


if __name__ == "__main__":
    main()


from abc import abstractmethod
from ..utils import decode_distribution
import numpy as np
import copy
from .utils import logging


class AMBase(object):
    def __init__(self, class_dict=[]):
        """ class_dict could be get from the data manager object """
        assert isinstance(class_dict, (list, tuple))
        self._class_dict = sorted(class_dict, key=lambda x: x['class_id'])
        self._class2label_all = {x['class_name']:x['class_id'] for x in class_dict}
        self._label2class_all = {x['class_id']:x['class_name'] for x in class_dict}

        self._all_classes = [x['class_name'] for x in self._class_dict]
        self._leaf_classes = self._get_leaf_classes(class_dict)

        self.set_class()

    @property
    def classnames(self):
        # set in self.set_class()
        return self._distribution_classes

    def clone(self):
        return copy.deepcopy(self)
    
    #
    # class manipulateion Methods
    #
    def _get_leaf_classes(self, class_dict):
        # 1) collect all parent classnames
        nonleaf_classes = set([x['parent'] for x in class_dict if x['parent'] is not None])
        leaf_classes = list()
        for x in self._class_dict:
            if x['class_name'] not in nonleaf_classes:
                leaf_classes.append(x['class_name'])
        return leaf_classes

    def _check_classes_conflict(self, classname_list):
        """ because we have parent classes, this function checks if there are parent class
            and its child class in the classname_list.
        """
        # loop over classname_list to check if one's parent is also in classname_list
        for classname in classname_list:
            parent = 'WHAT THE FUCK'
            child  = classname
            while parent is not None:
                classid_global = self._class2label_all[child]
                parent = self._class_dict[classid_global]['parent']
                if parent in classname_list:
                    raise RuntimeError("(WARNING) conflict class pair found: ({}, {})".format(child, parent))
                child = parent

    def _extend_distribution(self, distribution_global, exclude_classes=[]):
        """ extend the global distribution: fill parents with their maximal child """
        for child_id, child_score in enumerate(distribution_global):
            # trace back to root class
            parent = self._class_dict[child_id]['parent']
            while parent is not None:
                parent_id = self._class2label_all[parent]
                parent_score = distribution_global[parent_id]
                if parent not in exclude_classes:
                    distribution_global[parent_id] = max(child_score, parent_score)  # update parent score
                parent = self._class_dict[parent_id]['parent']   # update parent
                child_score = max(child_score, parent_score)     # update child_score
        return distribution_global

    @staticmethod
    def _add_ok_class(class_dict):
        """ Add OK class to class dict """
        __OK__ = 'OK'
        new_class_dict = copy.deepcopy(class_dict)
        for class_unit in new_class_dict:
            class_unit['class_id'] += 1
            assert class_unit['class_id'] != 0
            assert class_unit['class_name'] != __OK__
        OK_class_unit = {
            'class_id': 0,
            'class_name': __OK__,
            'parent': None
        }
        return [OK_class_unit] + new_class_dict



    @abstractmethod
    def set_class(self, arg):
        raise NotImplementedError()

    @abstractmethod
    def get_score(self, arg):
        raise NotImplementedError()

    #@logging
    def get_classid(self, arg):
        if None: pass
        elif isinstance(arg, dict):
            if 'instances' in arg:  # a record
                return [self.get_classid(inst) for inst in arg['instances']]
            else:                   # a instance
                inst = arg
                if 'class_id' in inst:
                    return inst['class_id']
                elif 'class_name' in inst:
                    classname = inst['class_name']
                    assert classname in self._class2label, "{} not in class dict: {}".format(classname, self._class2label)
                    return self._class2label[classname]
                elif 'distribution' in inst:
                    dist_all = decode_distribution(inst['distribution'])
                    dist_oin = [dist_all[ix] for ix in self._classid_of_interested]
                    return int(np.argmax(dist_oin))
                else:
                    raise RuntimeError()
        elif isinstance(arg, str):
            classname = arg
            assert classname in self._class2label, "{} not in class dict: {}".format(classname, self._class2label)
            return self._class2label[classname]
        elif isinstance(arg, list):
            return [self.get_classid(x) for x in arg]
        elif isinstance(arg, int):
            class_id = arg
            assert class_id in self._label2class
            return class_id
        else:
            raise TypeError("Argument Type Not Supported: {}".format(type(arg)))

    #@logging
    def get_classname(self, arg):
        res = self.get_classid(arg)
        if isinstance(res, list):
            return [self.get_classname(class_id) for class_id in res]
        else:
            class_id = res
            assert class_id in self._label2class
            return self._label2class[class_id]



from pathlib import Path
from ..utils import create_uuid
from ....utils_vvd import cv_rgb_imread
from ....utils_vvd import glob_recursively
from tqdm import tqdm
from .labelme import Labelme


def make_ok_labelme_json(image_path):
    if Path(image_path).exists():
        json_path = Path(image_path).with_suffix('.json')
        if not json_path.exists():
            uuid = create_uuid()
            realative_image_path = Path(image_path).name
            image = cv_rgb_imread(image_path)
            H, W = image.shape[:2]
            imaging = 2

            info = {
                'uuid': uuid,
                'image_path': realative_image_path,
                'height': H, 'width': W,
                'roi': [0, 0, H, W],  # roi in parent_image: xyxy
                'parent_uuid': uuid,
                "imaging": imaging,  # 1,2,3,4
            }

            labelme_info = Labelme(info, [])
            labelme_info.save_json(str(json_path))


def make_ok_labelme_for_dir(root_dir, img_suffix='png'):
    image_path_list = glob_recursively(root_dir, img_suffix)
    for image_path in tqdm(image_path_list):
        make_ok_labelme_json(image_path)